use Cwd qw( abs_path cwd );
use File::Basename;
use Getopt::Long;
use File::Spec::Unix;
use IPC::Open3;
use POSIX ":sys_wait_h"; 
use strict;

$|=1;

# Can contain as many options as needed and/or the test name, e.g. "--mysqld=--lock_wait_timeout=10 alias"
my $server_start_options  = "--mysqld=--log-bin=master-bin --mysqld=--log_output=TABLE --mysqld=--loose-performance_schema=0 --mysqld=--loose-lock_wait_timeout=3 alias";  

my $instances             = 1;        # Number of mysql-stress-test processes to run in parallel
my $server_run_pid        = 0;
my @stress_job_pids       = ();
my $abort_on_error        = 3;
my $top_pid               = $$;


my $opt_skip_server_start    = 0;
my $opt_skip_init            = 0;
my $opt_mysqltest            = undef;
my $opt_mysqladmin           = undef;
my $opt_port                 = 13100;
my $opt_help                 = 0;
my $opt_test_duration        = 600;
my $opt_threads_per_instance = 6;

&GetOptions ( 	
	"skip-server-start"  => \$opt_skip_server_start,
	"skip_server_start"  => \$opt_skip_server_start,
	"skip-init"          => \$opt_skip_init,
	"skip_init"          => \$opt_skip_init,
	"mysqltest=s"        => \$opt_mysqltest,
	"port=i"             => \$opt_port,
	"help"               => \$opt_help,
	"test-duration"      => \$opt_test_duration,
	"test_duration"      => \$opt_test_duration,
	"threads=i"          => \$opt_threads_per_instance,
);

if ( $opt_help ) 
{
	print 
"\
Usage: $0 [options] \
\
		The scripts starts MySQL server using MTR --start option, and then runs one or several \
		instances of mysql-stress-test with several threads, depending on a scenario. \
		Slave and replication might also be started if the scenario requires this. \
\
        Options:\n\
        --skip-server-start  The script won't start MySQL server, it's useful if you need to debug \ 
                             or on any other reason run the test on your running server. \
                             Please note that init job (pre-creating objects, if required) \
                             will still be executed by default. If you start tests on a running server \
                             several times, or on an existing vardir it might cause false-negatives. \
                             In this case, consider using --skip-init option. \
                             Also, if you skip starting server, nothing will be done for setting up replication. \
                             If it is needed, please make sure you configure it properly. \
\
        --skip-init          The script won't run init job, which creates objects using in stress tests. \
                             It is useful if you run the test several times on a running server, or start
                             server with a 'dirty' vardir (already containing objects from previous runs). \
\
        --mysqltest          The script attempts to find mysqltest binary in several folders. If the search \
                             fails, it will abort. In this case, please provide a full path to the binary \
                             using this option. \
\
        --port               Server port (for master server). If the script starts the server, this port \ 
                             will be used instead of default 13100. If server startup is skipped, it should be \
                             the port where your server is running (and, if the test requires slave, slave port \
                             should be equal master port + 1) \
\
        --test-duration      Test duration in seconds (especially useful if the test does not abort in the middle. \
\
        --threads            Number of threads per mysql-stress-test instance started by the test \
\
        --help               Print this help and exit. \n\
";
	exit;
}

if ( $opt_mysqltest ) {
	check_binary( $opt_mysqltest ) || die "mysqltest $opt_mysqltest does not exist or is not executable\n";
}
else {	
	$opt_mysqltest = find_mysqltest_binary();
	unless ( $opt_mysqltest ) {
		kill_all();
		die "Could not find mysqltest binary, please use --mysqltest option to provide full path\n";
	}
}

$ENV{MASTER_MYPORT} = $opt_port;
$ENV{SLAVE_MYPORT}  = $opt_port + 1;
my $mtr_build_thread = ( $opt_port - 10000 ) / 10;

unless ( $opt_skip_server_start )
{
	$server_run_pid = fork();
	die "Could not fork" unless defined $server_run_pid;

	unless ( $server_run_pid ) {
		my $cmd = "perl mysql-test-run.pl --mtr-build-thread=$mtr_build_thread --start $server_start_options";
		print "Starting server:\n$cmd\n";
		system( $cmd );	
		exit;
	}
}

check_if_can_connect() || die "ERROR: Could not connect to server, exiting\n";

unless ( $opt_skip_init )	
{
		my $cmd = "perl stress_test_basedir/mysql-stress-test.pl --server-host=127.0.0.1 --server-port=$opt_port "
			. "--stress-suite-basedir=stress_test_basedir "
			. "--stress-basedir=stress_test_basedir/logs/res_init/ "
			. "--server-logs-dir=stress_test_basedir/logs/log_init/ "
			. "--stress-init-file=stress_test_basedir/init.txt "
			. "--mysqltest=$opt_mysqltest "
			. "--log-error-details "
			. "--test-duration=$opt_test_duration "
			. "--server-user=root "
			. "--cleanup "
			. "--abort-on-error=3 ";
			
		system( "$cmd" );
		if ( ( $? >> 8 ) > 20 ) {
			kill_all();
			die "ERROR: Init job failed\n";
		}
}

for my $i ( 1..$instances )
{
	my $stress_pid = undef;
	$stress_pid = fork();
	unless ( defined $stress_pid ) {
		die "Could not fork for mysql-stress-test instance $i\n"; 
	}
	
	if ( $stress_pid ) {
		push @stress_job_pids, $stress_pid;
	}
	else {
		my $cmd = "perl stress_test_basedir/mysql-stress-test.pl --server-host=127.0.0.1 --server-port=$opt_port "
			. "--stress-suite-basedir=stress_test_basedir "
			. "--stress-basedir=stress_test_basedir/logs/res_run" . $i . "/ "
			. "--stress-tests-file=stress_test_basedir/stress" . $i . ".txt "
			. "--server-logs-dir=stress_test_basedir/logs/log_run" . $i . "/ "
			. "--mysqltest=../bin/mysqltest "
			. "--log-error-details "
			. "--threads=$opt_threads_per_instance "
			. "--test-duration=$opt_test_duration "
			. "--server-user=root "
			. "--cleanup "
			. "--abort-on-error=$abort_on_error ";
		
		system( "$cmd" );
		exit $?;
	}
} 

while ( 1 )
{
	foreach ( @stress_job_pids )
	{
		waitpid( $_, WNOHANG );
		if ( $? > -1 ) {
			kill_all();
			exit;
		} 
	}	
	sleep( 2 );
}
	
sub find_mysqltest_binary 
{
	my $name = 'mysqltest';
	
	my @folders = ();
	my $start_folder = dirname( cwd() );
	
	@folders = qw( sql libexec client client_release client_debug client/release 
		client/relwithdebinfo client/debug sql/release sql/relwithdebinfo sql/debug sbin bin 
		share/mysql sql/share share lib/plugin plugin );

	foreach my $d ( @folders ) 
	{
		my $f = File::Spec::Unix->catfile( $start_folder, $d, $name );
		return $f if check_binary( $f );
	}
	# Last resort -- maybe it's on path
	return if check_binary( $name );
	return undef;	
}

sub check_binary 
{
	my $binary = shift;
	my $pid = open3( undef, *OUT, *ERR, $binary, "--version" );
	while (<OUT>) {}
	while (<ERR>) {}
	close(OUT);
	close(ERR);
	waitpid $pid, 0;
	return ! $?;
}

sub check_if_can_connect
{
	print "Trying to connect to server at 127.0.0.1:$opt_port...\n";
	if ( open( CONN, " | $opt_mysqltest --user=root --port=$opt_port --host=127.0.0.1 --max-connect-retries=1000" ) )
	{
		print CONN "--echo # mysqltest was able to connect to server\n--exit\n\n";
		close( CONN );
	}	
	return ! $?;
}

sub kill_all
{
	foreach ( @stress_job_pids ) {
		kill 'KILL', $_;
	}
	kill 'INT', $server_run_pid if $server_run_pid;
	kill 'TERM', $top_pid;
}
